local BtWQuestsDatabase = BtWQuestsDatabase
local BtWQuestsCharacters = BtWQuestsCharacters

local BTWQUESTS_CATEGORY_ITEM_WIDTH = 174
local BTWQUESTS_CATEGORY_ITEM_HEIGHT = 96
local BTWQUESTS_CATEGORY_ITEM_PADDING = 12
local BTWQUESTS_CATEGORY_NUM_ITEMS_PER_ROW = 4

local BTWQUESTS_CHAIN_ITEM_WIDTH = 174 / 2
local BTWQUESTS_CHAIN_ITEM_HEIGHT = 80

BtWQuestsMixin = {}
function BtWQuestsMixin:GetCharacter()
    return self.Character
end
function BtWQuestsMixin:SelectCharacter(name, realm)
    local key
    if realm == nil then
        key = name
    else
        key = name .. "-" .. realm
    end

    local character = BtWQuestsCharacters:GetCharacter(key)
    if character ~= nil then
        self.Character = key
        -- UIDropDownMenu_SetSelectedValue(self.CharacterDropDown, key)
        UIDropDownMenu_SetText(self.CharacterDropDown, key)

        self:Refresh()
    end
end

function BtWQuestsMixin:GetExpansion()
    return self.expansionID
end
function BtWQuestsMixin:SetExpansion(id)
    self.expansionID = tonumber(id)
    self.categoryID = nil
    self.chainID = nil
end
function BtWQuestsMixin:SelectExpansion(id, scrollTo, noHistory)
    if not noHistory then
        self:UpdateCurrentHistory()
    end

    self:SetExpansion(id)
    
    NavBar_Reset(self.navBar)
    
    local expansion = self:GetExpansion()
    if expansion and BtWQuestsDatabase:HasExpansion(expansion) then
	    local questSelect = self.Category;
        local tierData = BtWQuestsDatabase.Expansions[expansion].background;
        if tierData.atlas then
            questSelect.Background:SetAtlas(tierData.atlas);    
        else
            questSelect.Background:SetTexture(tierData.texture);
        end
        UIDropDownMenu_SetText(questSelect.ExpansionDropDown, BtWQuestsDatabase:GetExpansionByID(expansion));
    end

    self:DisplayCurrentExpansion(scrollTo)

    if not noHistory then
        self:AddCurrentToHistory()
    end
end


function BtWQuestsMixin:GetCategory()
    return self.categoryID
end
function BtWQuestsMixin:SetCategory(id)
    if id == nil then
        self.categoryID = nil
        self.chainID = nil
    else
        self.categoryID = tonumber(id)
        self.chainID = nil
        self.expansionID = select(4, BtWQuestsDatabase:GetCategoryByID(self.categoryID))
    end
end
function BtWQuestsMixin:SelectCategory(id, scrollTo, noHistory)
    if not noHistory then
        self:UpdateCurrentHistory()
    end
    
    self:SetCategory(id)
    
    self.navBar:SetCategory(id)
    
    local expansion = self:GetExpansion()
    if expansion and BtWQuestsDatabase:HasExpansion(expansion) then
	    local questSelect = self.Category;
        local tierData = BtWQuestsDatabase.Expansions[expansion].background;
        if tierData.atlas then
            questSelect.Background:SetAtlas(tierData.atlas);    
        else
            questSelect.Background:SetTexture(tierData.texture);
        end
        UIDropDownMenu_SetText(questSelect.ExpansionDropDown, BtWQuestsDatabase:GetExpansionByID(expansion));
    end

    self:DisplayCurrentCategory(scrollTo)

    if not noHistory then
        self:AddCurrentToHistory()
    end
end


function BtWQuestsMixin:GetChain()
    return self.chainID
end
function BtWQuestsMixin:SetChain(id)
    self.chainID = tonumber(id)
    self.expansionID, self.categoryID = select(4, BtWQuestsDatabase:GetChainByID(self.chainID))
end
function BtWQuestsMixin:SelectChain(id, scrollTo, noHistory)
    if not noHistory then
        self:UpdateCurrentHistory()
    end
    
    self:SetChain(id)
    
    self.navBar:SetChain(id)
    
    local expansion = self:GetExpansion()
    if expansion and BtWQuestsDatabase:HasExpansion(expansion) then
	    local questSelect = self.Category;
        local tierData = BtWQuestsDatabase.Expansions[expansion].background;
        if tierData.atlas then
            questSelect.Background:SetAtlas(tierData.atlas);    
        else
            questSelect.Background:SetTexture(tierData.texture);
        end
        UIDropDownMenu_SetText(questSelect.ExpansionDropDown, BtWQuestsDatabase:GetExpansionByID(expansion));
    end

    self:DisplayCurrentChain(scrollTo)
    
    if not noHistory then
        self:AddCurrentToHistory()
    end
end


function BtWQuestsMixin:SelectFromLink(link, scrollTo)
    local _, _, color, type, text, name = string.find(link, "|cff(%x*)|H([^:]+):([^|]+)|h%[([^%[%]]*)%]|h|r")
    if not color then
        _, _, type, text = string.find(link, "([^:]+):(.+)")
    end
    
    assert(type == "quest" or type == "btwquests")
    
    if type == "quest" then
        local _, _, id = string.find(text, "^(%d+):")

        id = tonumber(id)
        
        local questLogIndex = GetQuestLogIndexByID(id);
        if questLogIndex > 0 then
            if IsQuestComplete(id) and GetQuestLogIsAutoComplete(questLogIndex) then
                AutoQuestPopupTracker_RemovePopUp(id);
                ShowQuestComplete(questLogIndex);
                
                return true
            else
                QuestMapFrame_OpenToQuestDetails(id)
                
                return true
            end
        end
    elseif type == "btwquests" then
        local _, _, subtype, id = string.find(text, "^([^:]*):(%d+)")
        
        assert(subtype == "expansion" or subtype == "category" or subtype == "chain")
        
        if subtype == "expansion" then
            self:SelectExpansion(id, scrollTo)
            
            return true
        elseif subtype == "category" then
            self:SelectCategory(id, scrollTo)
            
            return true
        elseif subtype == "chain" then
            self:SelectChain(id, scrollTo)
            
            return true
        end
    end
    
    return false
end

function BtWQuestsMixin:SelectItem(item, scrollTo)
    if item.type == "expansion" then
        self:SelectExpansion(item.id, scrollTo or item.scrollTo)
    elseif item.type == "category" then
        self:SelectCategory(item.id, scrollTo or item.scrollTo)
    elseif item.type == "chain" then
        self:SelectChain(item.id, scrollTo or item.scrollTo)
    end
end

function BtWQuestsMixin:SelectFromHistory()
    local item = self.History[self.HistoryIndex]
    
    if item.type == "chain" then
        self:SelectChain(item.id, item.scrollTo, true)
    elseif item.type == "category" then
        self:SelectCategory(item.id, item.scrollTo, true)
    elseif item.type == "expansion" then
        self:SelectExpansion(item.id, item.scrollTo, true)
    end
end

function BtWQuestsMixin:Back()
    if self.HistoryIndex > 1 then
        self:UpdateCurrentHistory()
        
        self.HistoryIndex = self.HistoryIndex - 1
        
        self:SelectFromHistory()
    
        self:UpdateHistoryButtons()
    end
end

function BtWQuestsMixin:Forward()
    if self.HistoryIndex < #self.History then
        self:UpdateCurrentHistory()
        
        self.HistoryIndex = self.HistoryIndex + 1
        
        self:SelectFromHistory()
    
        self:UpdateHistoryButtons()
    end
end

function BtWQuestsMixin:Here()
    local item = BtWQuestsDatabase:GetMapItemByID(C_Map.GetBestMapForUnit("player"), BtWQuestsCharacters:GetCharacter(self:GetCharacter()))

    if item == nil then
        self.Here:Disable()
    else
        self:SelectItem(item)
    end
end

function BtWQuestsMixin:ZoomOut()
    self:Back()
    
    self.Tooltip:Hide();
end

function BtWQuestsMixin:GetCurrentView()
    if self.Chain:IsShown() then
        return {
            type = "chain",
            id = self:GetChain(),
            scrollTo = {
                type = "coords",
                x = self.Chain.Scroll:GetHorizontalScroll(),
                y = self.Chain.Scroll:GetVerticalScroll(),
            }
        };
    elseif self:GetCategory() ~= nil then
        return {
            type = "category",
            id = self:GetCategory(),
            scrollTo = {
                type = "coords",
                x = self.Category.Scroll:GetHorizontalScroll(),
                y = self.Category.Scroll:GetVerticalScroll(),
            }
        };
    else
        return {
            type = "expansion",
            id = self:GetExpansion(),
            scrollTo = {
                type = "coords",
                x = self.Category.Scroll:GetHorizontalScroll(),
                y = self.Category.Scroll:GetVerticalScroll(),
            }
        };
    end
end

function BtWQuestsMixin:AddCurrentToHistory()
    local last = self.History[self.HistoryIndex]
    local current = self:GetCurrentView()
    
    if last == nil or current.type ~= last.type or current.id ~= last.id then
        self.HistoryIndex = self.HistoryIndex + 1

        while self.History[self.HistoryIndex] do
            table.remove(self.History, self.HistoryIndex)
        end

        table.insert(self.History, current);
    end
    
    self:UpdateHistoryButtons()
end

function BtWQuestsMixin:UpdateCurrentHistory()
    self.History[self.HistoryIndex] = self:GetCurrentView();
end

function BtWQuestsMixin:UpdateHistoryButtons()
    self.NavBack:SetEnabled(self.HistoryIndex > 1)
    self.NavForward:SetEnabled(self.HistoryIndex < #self.History)
end

function BtWQuestsMixin:UpdateHereButton()
    self.NavHere:SetEnabled(BtWQuestsDatabase:GetMapItemByID(C_Map.GetBestMapForUnit("player"), BtWQuestsCharacters:GetCharacter(self:GetCharacter())) ~= nil)
end

function BtWQuestsMixin:DisplayCurrentCategory(scrollTo)
	self.categoryItemPool:ReleaseAll();
	local questSelect = self.Category;
    
    local character = BtWQuestsCharacters:GetCharacter(self:GetCharacter())
    self.Tooltip.character = character
    
	self.Chain:Hide();
	questSelect:Show();
    
	local scrollFrame = questSelect.Scroll.Child;
    local scrollToButton
    
    local startX = 12
    local startY = -10
	local i = 1;
	local index = 1;
    local itemType, id, name, hidden, category, expansion, buttonImage, onClick, onEnter, onLeave, userdata = BtWQuestsDatabase:GetCategoryItemByIndex(self:GetCategory(), i, character)
    while id do
        if not hidden then
            local categoryButton = self.categoryItemPool:Acquire();

            local x = startX + mod(index - 1, BTWQUESTS_CATEGORY_NUM_ITEMS_PER_ROW) * (BTWQUESTS_CATEGORY_ITEM_WIDTH + BTWQUESTS_CATEGORY_ITEM_PADDING)
            local y = startY - floor((index - 1) / BTWQUESTS_CATEGORY_NUM_ITEMS_PER_ROW) * (BTWQUESTS_CATEGORY_ITEM_HEIGHT + BTWQUESTS_CATEGORY_ITEM_PADDING)
            
            categoryButton:SetPoint("TOPLEFT", x,  y);

            categoryButton.name:SetText(name);
            categoryButton.bgImage:SetTexture(buttonImage);
            categoryButton.AciveTexture:SetShown(itemType == "chain" and BtWQuestsDatabase:IsChainActive(id, character) or false)
            if itemType == "chain" then
                categoryButton.Tick:SetShown(BtWQuestsDatabase:IsChainCompleted(id, character))
            else
                categoryButton.Tick:SetShown(BtWQuestsDatabase:IsCategoryCompleted(id, character))
            end
            
            categoryButton.type = itemType;
            categoryButton.id = id;
            categoryButton.userdata = userdata;
            categoryButton.Tooltip = self.Tooltip
            categoryButton:SetScript("OnClick", onClick)
            categoryButton:SetScript("OnEnter", onEnter)
            categoryButton:SetScript("OnLeave", onLeave)
            
            categoryButton:Show();

            if type(scrollTo) == "number" and index == scrollTo then
                scrollToButton = categoryButton
            elseif type(scrollTo) == "table" and scrollTo.type ~= "coords" and BtWQuestsDatabase:CompareCategoryItemByIndex(self:GetCategory(), index, scrollTo) then
                scrollToButton = categoryButton
            end

            index = index + 1;
        end
        
        i = i + 1;
        itemType, id, name, hidden, category, expansion, buttonImage, onClick, onEnter, onLeave, userdata = BtWQuestsDatabase:GetCategoryItemByIndex(self:GetCategory(), i, character)
    end

    if type(scrollTo) == "table" and scrollTo.type == "coords" then
        questSelect.Scroll:UpdateScrollChildRect()
        -- questSelect.Scroll:SetHorizontalScroll(scrollTo.x)
        questSelect.Scroll:SetVerticalScroll(scrollTo.y)
    elseif scrollTo ~= false then
        if scrollToButton then
            questSelect.Scroll:UpdateScrollChildRect()
            questSelect.Scroll:SetVerticalScroll(-select(5, scrollToButton:GetPoint("TOP")) - (questSelect.Scroll:GetHeight()/2) + 24)
        end
    end
    
    ShowUIPanel(BtWQuests)
end

function BtWQuestsMixin:DisplayCurrentExpansion(scrollTo)
	self.categoryItemPool:ReleaseAll();
    local questSelect = self.Category;
    
    local character = BtWQuestsCharacters:GetCharacter(self:GetCharacter())
    self.Tooltip.character = character
    
	self.Chain:Hide();
	questSelect:Show();
    
	local scrollFrame = questSelect.Scroll.Child;
    local scrollToButton
    
    local startX = 12
    local startY = -10
	local i = 1;
	local index = 1;
    local itemType, id, name, hidden, category, expansion, buttonImage, onClick, onEnter, onLeave, userdata = BtWQuestsDatabase:GetExpansionItemByIndex(self:GetExpansion(), i, character);
    while id do
        if not hidden then
            local categoryButton = self.categoryItemPool:Acquire();

            local x = startX + mod(index - 1, BTWQUESTS_CATEGORY_NUM_ITEMS_PER_ROW) * (BTWQUESTS_CATEGORY_ITEM_WIDTH + BTWQUESTS_CATEGORY_ITEM_PADDING)
            local y = startY - floor((index - 1) / BTWQUESTS_CATEGORY_NUM_ITEMS_PER_ROW) * (BTWQUESTS_CATEGORY_ITEM_HEIGHT + BTWQUESTS_CATEGORY_ITEM_PADDING)
            
            categoryButton:SetPoint("TOPLEFT", x,  y);

            categoryButton.name:SetText(name);
            categoryButton.bgImage:SetTexture(buttonImage);
            categoryButton.AciveTexture:SetShown(itemType == "chain" and BtWQuestsDatabase:IsChainActive(id, character) or false)
            if itemType == "chain" then
                categoryButton.Tick:SetShown(BtWQuestsDatabase:IsChainCompleted(id, character))
            else
                categoryButton.Tick:SetShown(BtWQuestsDatabase:IsCategoryCompleted(id, character))
            end
            
            categoryButton.type = itemType;
            categoryButton.id = id;
            categoryButton.userdata = userdata;
            categoryButton.Tooltip = self.Tooltip;
            categoryButton:SetScript("OnClick", onClick)
            categoryButton:SetScript("OnEnter", onEnter)
            categoryButton:SetScript("OnLeave", onLeave)
            
            categoryButton:Show();

            if type(scrollTo) == "number" and index == scrollTo then
                scrollToButton = categoryButton
            elseif type(scrollTo) == "table" and scrollTo.type ~= "coords" and BtWQuestsDatabase:CompareExpansionItemByIndex(self:GetExpansion(), index, scrollTo) then
                scrollToButton = categoryButton
            end

            index = index + 1;
        end
        
        i = i + 1;
        itemType, id, name, hidden, category, expansion, buttonImage, onClick, onEnter, onLeave, userdata = BtWQuestsDatabase:GetExpansionItemByIndex(self:GetExpansion(), i, character);
    end

    if type(scrollTo) == "table" and scrollTo.type == "coords" then
        questSelect.Scroll:UpdateScrollChildRect()
        -- questSelect.Scroll:SetHorizontalScroll(scrollTo.x)
        questSelect.Scroll:SetVerticalScroll(scrollTo.y)
    elseif scrollTo ~= false then
        if scrollToButton then
            questSelect.Scroll:UpdateScrollChildRect()
            questSelect.Scroll:SetVerticalScroll(-select(5, scrollToButton:GetPoint("TOP")) - (questSelect.Scroll:GetHeight()/2) + 24)
        end
    end
    
    ShowUIPanel(BtWQuests)
end

function BtWQuestsMixin:DisplayCurrentChain(scrollTo, zoom)
	local chain = self.Chain;
    
	self.Category:Hide();
    chain:Show();
    
    chain.Scroll:SetCharacter(self:GetCharacter())
    chain.Scroll:SetChain(self:GetChain(), scrollTo, zoom)

    ShowUIPanel(self)
end

function BtWQuestsMixin:UpdateCurrentChain()
    local chain = self.Chain
    
    if chain:IsShown() then
        chain.Scroll:Update()
    end
end

local function ChainItemPool_HideAndClearAnchors(framePool, frame)
    FramePool_HideAndClearAnchors(framePool, frame)
    
    if frame.backgroundLinePool then
        frame.backgroundLinePool:ReleaseAll();
    end
end

tinsert(UISpecialFrames, "BtWQuestsFrame") 
function BtWQuestsMixin:OnLoad()
    self:RegisterEvent("ADDON_LOADED")
    self:RegisterEvent("QUEST_LOG_UPDATE")

    self:RegisterEvent("ZONE_CHANGED")
    self:RegisterEvent("ZONE_CHANGED_INDOORS")
    self:RegisterEvent("ZONE_CHANGED_NEW_AREA")
    
    self:SetAttribute("UIPanelLayout-defined", true)
    self:SetAttribute("UIPanelLayout-enabled", true)
    self:SetAttribute("UIPanelLayout-area", "left")
    self:SetAttribute("UIPanelLayout-pushable", 0)
    self:SetAttribute("UIPanelLayout-width", 830)
    self:SetAttribute("UIPanelLayout-whileDead", true)
    
	self.TitleText:SetText(BTWQUESTS_QUEST_JOURNAL);
    SetPortraitToTexture(self.portrait, "Interface\\QuestFrame\\UI-QuestLog-BookIcon");
    
    self.TopBorder:SetPoint("TOPLEFT", self.NavBorderRight, "TOPRIGHT", 0, 0);
    self.TopBorder:SetPoint("TOPRIGHT", self.CharacterDropDown, "TOPLEFT", 0, 0);
    
    self.Tooltip:SetBackdropBorderColor(TOOLTIP_DEFAULT_COLOR.r, TOOLTIP_DEFAULT_COLOR.g, TOOLTIP_DEFAULT_COLOR.b);
    self.Tooltip:SetBackdropColor(TOOLTIP_DEFAULT_BACKGROUND_COLOR.r, TOOLTIP_DEFAULT_BACKGROUND_COLOR.g, TOOLTIP_DEFAULT_BACKGROUND_COLOR.b);
    
    self.categoryItemPool = CreateFramePool("BUTTON", self.Category.Scroll.Child, "BtWQuestsCategoryButtonTemplate");
    self.chainItemPool = CreateFramePool("BUTTON", self.Chain.Scroll.Child, "BtWQuestsChainItemButtonTemplate", ChainItemPool_HideAndClearAnchors);
    
    self.HistoryIndex = 0
    self.History = {}
    -- self:AddCurrentToHistory()
    
	-- LDB launcher
	local LDB = LibStub and LibStub("LibDataBroker-1.1", true)
	if LDB then
		BtWQuestsLauncher = LDB:NewDataObject("BtWQuests", {
			type = "launcher",
            label = "BtWQuests",
			icon = "Interface\\QuestFrame\\UI-QuestLog-BookIcon",
			OnClick = function(clickedframe, button)
				ToggleFrame(BtWQuestsFrame)
			end,
		})
	end
end

function BtWQuestsMixin:OnEvent(event, ...)
    if event == "ADDON_LOADED" then
        if ... == "BtWQuests" then
            hooksecurefunc("QuestObjectiveTracker_OnOpenDropDown", function (self)
                BtWQuests_AddOpenChainMenuItem(self, self.activeFrame.id)
            end)
            hooksecurefunc(QuestMapQuestOptionsDropDown, "initialize", function (self)
                BtWQuests_AddOpenChainMenuItem(self, self.questID)
            end)

            WorldMapFrame:AddDataProvider(CreateFromMixins(BtWQuestsQuestDataProviderMixin));
        end
    elseif event == "QUEST_LOG_UPDATE" then
        BtWQuestsCharacters:UpdatePlayer()
    elseif event == "ZONE_CHANGED" or event == "ZONE_CHANGED_INDOORS" or event == "ZONE_CHANGED_NEW_AREA" then
        if self:IsShown() then
            self:UpdateHereButton()
        end
    end
end

function BtWQuestsMixin:Refresh()
    if self:GetChain() ~= nil then
        self:DisplayCurrentChain()
    elseif self:GetCategory() ~= nil then
        self:DisplayCurrentCategory()
    elseif self:GetExpansion() ~= nil then
        self:DisplayCurrentExpansion()
    end
end

function BtWQuestsMixin:OnShow()
    PlaySound(SOUNDKIT.IG_CHARACTER_INFO_OPEN);
    
    self:UpdateHereButton()

    if not self.initialized then
    --     local questSelect = self.Category;
    --     UIDropDownMenu_SetWidth(questSelect.ExpansionDropDown, 156);
    --     UIDropDownMenu_JustifyText(questSelect.ExpansionDropDown, "LEFT");
    --     UIDropDownMenu_Initialize(questSelect.ExpansionDropDown, BtWQuestsExpansionDropDown_Initialize);
    
    --     local expansion = BtWQuests_GetCurrentExpansion()
    --     if expansion and BtWQuests_Expansions[expansion] then
    --         local tierData = BtWQuests_Expansions[expansion].background;
    --         if tierData.atlas then
    --             questSelect.bg:SetAtlas(tierData.atlas);    
    --         else
    --             questSelect.bg:SetTexture(tierData.texture);
    --         end
    --         UIDropDownMenu_SetText(questSelect.ExpansionDropDown, BtWQuests_GetExpansionInfo(BtWQuests_GetCurrentExpansion()));
    --     end

        -- UIDropDownMenu_SetWidth(self.CharacterDropDown, 156, 23 * 2);
        -- UIDropDownMenu_Initialize(self.CharacterDropDown, BtWQuestsCharacterDropDown_Initialize);

        -- self.CharacterDropDown.xOffset = 8
        -- self.CharacterDropDown.yOffset = 15
        self:SelectCharacter(UnitName("player"), GetRealmName())

        if self:GetExpansion() == nil then -- Not guessed/set an expansion yet
            self:SelectExpansion(BtWQuestsDatabase:GuessExpansion(self.Character))
        end
    --     local name = UnitName("player")
    --     local realm = GetRealmName()
    --     self.CharacterDropDown.xOffset = 8
    --     self.CharacterDropDown.yOffset = 15
    --     UIDropDownMenu_SetText(self.CharacterDropDown, name .. "-" .. realm)

    --     local homeData = {
    --         name = HOME,
    --         OnClick = function()
    --             BtWQuestsNav_UpdateCurrentHistory()
        
    --             BtWQuests_SetCurrentCategory(nil)
                
    --             BtWQuests_DisplayCurrentCategory()
    
    --             BtWQuestsNav_AddCurrentToHistory()
    --         end,
    --     }
    --     NavBar_Initialize(self.navBar, "NavButtonTemplate", homeData, self.navBar.home, self.navBar.overflow);

        self.initialized = true
    else
        if self:GetChain() ~= nil then
            self:UpdateCurrentChain()
        elseif self:GetCategory() ~= nil then
            self:DisplayCurrentCategory()
        elseif self:GetExpansion() ~= nil then
            self:DisplayCurrentExpansion()
        end
    end
end

function BtWQuestsMixin:OnHide(self)
	PlaySound(SOUNDKIT.IG_CHARACTER_INFO_CLOSE);
end

function BtWQuestsChainFrame_OnShow(self)
    self:RegisterEvent("QUEST_ACCEPTED")
    self:RegisterEvent("QUEST_AUTOCOMPLETE")
    self:RegisterEvent("QUEST_COMPLETE")
    self:RegisterEvent("QUEST_FINISHED")
    self:RegisterEvent("QUEST_TURNED_IN")
    self:RegisterEvent("QUEST_LOG_CRITERIA_UPDATE")
    self:RegisterEvent("QUEST_WATCH_LIST_CHANGED")
    self:RegisterEvent("QUEST_WATCH_UPDATE")
end

function BtWQuestsChainFrame_OnHide(self)
    self:UnregisterAllEvents()
end

function BtWQuestsChainFrame_OnEvent(self, ...)
    self:GetParent():UpdateCurrentChain()
end

function BtWQuestsChainFrameScrollFrame_OnUpdate(self)
    local mouseX, mouseY = GetCursorPosition()
    local scale = self:GetEffectiveScale()
    mouseX, mouseY = mouseX / scale, mouseY / scale
    
    local maxXScroll, maxYScroll = self:GetHorizontalScrollRange(), self:GetVerticalScrollRange()
    
    mouseX = min(max(mouseX - self.mouseX + self.scrollX, 0), maxXScroll)
    mouseY = min(max(mouseY - self.mouseY + self.scrollY, 0), maxYScroll)
    
    self:SetHorizontalScroll(mouseX)
    self:SetVerticalScroll(mouseY)
end

-- [[ Chain ]]
function BtWQuestsChainItemPool_HideAndClearAnchors(framePool, frame)
    FramePool_HideAndClearAnchors(framePool, frame)
    
    frame.linePool:ReleaseAll()
    
    frame.IsNextAnim:Stop()
    frame.previousButtons = nil
    frame.status = nil
end

BtWQuestsChainItemMixin = {}
function BtWQuestsChainItemMixin:OnLoad()
    self.linePool = CreateFramePool("FRAME", self:GetParent(), "BtWQuestsLineTemplate");
    self:RegisterForDrag("LeftButton")
end
function BtWQuestsChainItemMixin:Set(item)
    self.item = item

    self.ForgottenAnim:Stop()

    local status = item:GetStatus()
    if item:IsBreadcrumb() and item:HasConnections() then
        local completed = false
        local index = 1
        local connection = item:GetConnection(index)
        while connection do
            if not connection:GetSkip() and connection:GetVisible() and connection:GetStatus() ~= nil then
                completed = true
                break
            end

            index = index + 1
            connection = item:GetConnection(index)
        end
        
        if completed then
            status = "complete"
        end
    end

    self.name:SetText(item:GetName())

    local tagID = item:GetTagID()
    if tagID and QUEST_TAG_TCOORDS[tagID] then
        self.TagTexture:SetTexCoord(unpack(QUEST_TAG_TCOORDS[tagID]))
        self.TagTexture:Show()

        if difficulty then
            self.TagTexture:SetPoint("BOTTOMRIGHT", -10, 6)
        else
            self.TagTexture:SetPoint("BOTTOMRIGHT", -10, 16)
        end
    end

    local difficulty = item:GetDifficulty()
    self.HeroicTexture:SetShown(difficulty == "heroic" or difficulty == "normal" or difficulty == "lfr")
    self.MythicTexture:SetShown(difficulty == "mythic")

    local atlas = item:GetAtlas()
    if atlas ~= nil then
        self.Icon:SetAtlas(atlas)
    end

    if status == "complete" then
        self.name:SetAlpha(0.5)
    else
        self.name:SetAlpha(1)
        -- self.ForgottenAnim:Play()
    end
    
    if status == "active" then
        self.ActiveTexture:Show()
        self.ActiveAnim:Play()
    else
        self.ActiveTexture:Hide()
    end
    
    self.Tick:SetShown(status == "complete")

    self:SetShown(item:GetVisible())
    self.status = status
end
function BtWQuestsChainItemMixin:Update(item)
    local status = item:GetStatus()
    if item:IsBreadcrumb() and item:HasConnections() then
        local completed = false
        local index = 1
        local connection = item:GetConnection(index)
        while connection do
            if not connection:GetSkip() and connection:GetVisible() and connection:GetStatus() ~= nil then
                completed = true
                break
            end

            index = index + 1
            connection = item:GetConnection(index)
        end
        
        if completed then
            status = "complete"
        end
    end

    self.name:SetText(item:GetName())
    
    if self.status ~= status then
        if status == "complete" then
            self.ForgottenAnim:Play()
        end
    
        if status == "active" then
            self.ActiveTexture:Show()
            self.ActiveAnim:Play()
        else
            self.ActiveTexture:Hide()
        end
    end
    
    self.Tick:SetShown(status == "complete")
    
    self:SetShown(item:GetVisible())
    self.status = status
end
function BtWQuestsChainItemMixin:OnDragStart()
    self:GetParent():GetParent():OnDragStart()
end
function BtWQuestsChainItemMixin:OnDragStop()
    self:GetParent():GetParent():OnDragStop()
end
function BtWQuestsChainItemMixin:OnClick()
    return self.item:OnClick(self.item, self, self:GetParent():GetParent(), self:GetParent():GetParent():GetTooltip())
end
function BtWQuestsChainItemMixin:OnEnter()
    return self.item:OnEnter(self.item, self, self:GetParent():GetParent(), self:GetParent():GetParent():GetTooltip())
end
function BtWQuestsChainItemMixin:OnLeave()
    return self.item:OnLeave(self.item, self, self:GetParent():GetParent(), self:GetParent():GetParent():GetTooltip())
end

BtWQuestsChainMixin = {}
function BtWQuestsChainMixin:OnLoad()
    ScrollFrame_OnLoad(self)
    self.itemPool = CreateFramePool("BUTTON", self.Child, "BtWQuestsChainItemTemplate", BtWQuestsChainItemPool_HideAndClearAnchors);
    self:RegisterForDrag("LeftButton")
end
function BtWQuestsChainMixin:OnDragStart()
    self.scrollX, self.scrollY = self:GetHorizontalScroll(), self:GetVerticalScroll()
    self.mouseX, self.mouseY = GetCursorPosition()
    
    local scale = self:GetScrollChild():GetEffectiveScale()
    self.mouseX, self.mouseY = self.mouseX / scale, self.mouseY / scale

    self:SetScript("OnUpdate", BtWQuestsChainMixin.OnDrag)
end
function BtWQuestsChainMixin:OnDragStop()
    self:SetScript("OnUpdate", nil)
end
function BtWQuestsChainMixin:OnDrag()
    local mouseX, mouseY = GetCursorPosition()
    local scale = self:GetEffectiveScale()
    mouseX, mouseY = mouseX / scale, mouseY / scale
    
    local maxXScroll, maxYScroll = self:GetHorizontalScrollRange(), self:GetVerticalScrollRange()
    
    mouseX = min(max(mouseX - self.mouseX + self.scrollX, 0), maxXScroll)
    mouseY = min(max(mouseY - self.mouseY + self.scrollY, 0), maxYScroll)
    
    self:SetHorizontalScroll(mouseX)
    self:SetVerticalScroll(mouseY)
end
function BtWQuestsChainMixin:GetZoom()
    return 1
    -- return self:GetScrollChild():GetScale()
end
function BtWQuestsChainMixin:SetZoom(value)
    -- return self:GetScrollChild():SetScale(value)
end
function BtWQuestsChainMixin:SetCharacter(name, realm)
    if realm ~= nil then
        name = name .. "-" .. realm
    end

    if type(name) == "string" then
        self.Character = BtWQuestsCharacters:GetCharacter(name)
    else
        self.Character = name
    end
end
function BtWQuestsChainMixin:GetCharacter()
    return self.Character or BtWQuestsCharacters:GetPlayer()
end
function BtWQuestsChainMixin:GetTooltip()
    return self:GetParent():GetParent().Tooltip or BtWQuestsTooltip
end
function BtWQuestsChainMixin:SelectFromLink(...)
    return self:GetParent():GetParent():SelectFromLink(...)
end
function BtWQuestsChainMixin:SetChain(chainID, scrollTo, zoom)
    self.chainID = chainID

    self.itemPool:ReleaseAll()
    self.ItemButtons = {}

    self.scrollTo = scrollTo
    self.scrollToButton = nil
    self.scrollToButtonAside = nil

    if zoom ~= nil then
        self:SetZoom(zoom)
    end

    local numItems = self:AddButtons(chainID, 0, 0)

    local buttons = self.ItemButtons[chainID]
    local temp = numItems
    while temp > 0 and (not buttons[temp] or not buttons[temp]:IsShown()) do
        temp = temp - 1
    end

    if temp > 0 then
        self:GetScrollChild().Bottom:SetPoint("TOP", 0, select(5, buttons[temp]:GetPoint("TOP")) - 23 - (self:GetHeight()/2))
    end

    self:UpdateScrollChildRect()
    if type(scrollTo) == "table" and scrollTo.type == "coords" then
        local scale = self:GetZoom()
        -- self:SetHorizontalScroll(scrollTo.x / scale)
        self:SetVerticalScroll(scrollTo.y / scale)
    elseif scrollTo ~= false then
        if self.scrollToButton == nil then
            self.scrollToButton = self.scrollToButtonAside
        end
        
        if self.scrollToButton == nil then
            local buttons = self.ItemButtons[chainID]
            local index = 1
            while buttons[index] and not buttons[index]:IsShown() do
                index = index + 1
            end
            
            self.scrollToButton = buttons[index]
        end
        
        if self.scrollToButton then
            local scale = self:GetZoom()
            local x, y = select(4, self.scrollToButton:GetPoint("TOP"))
            
            -- self:SetHorizontalScroll(x - (self:GetWidth() / scale) / 2)
            self:SetVerticalScroll(0 - y - (self:GetHeight() / scale) / 2 + 24)
        end
    end
end
function BtWQuestsChainMixin:AddButtons(chainID, xOffset, yOffset, asideOverride)
    local character = self:GetCharacter()

    self.ItemButtons[chainID] = {}
    local buttons = self.ItemButtons[chainID]

    local chain = BtWQuestsDatabase:GetChain(chainID, character)

    local index = 1
    local item = chain:GetItem(index)
    while item do
        if not item:GetSkip() then
            local itemButton = buttons[index] or self.itemPool:Acquire();
            buttons[index] = itemButton

            itemButton:Set(item)

            -- Check if the item should be next
            itemButton.IsNextAnim:Stop()
            if itemButton.status == nil and itemButton.previousButtons ~= nil then
                local available = true
                for i,previous in ipairs(itemButton.previousButtons) do
                    if previous.status ~= "complete" then
                        available = false
                        break
                    end
                end

                if available then
                    itemButton.ActiveTexture:Show()
                    itemButton.IsNextAnim:Play()
                end

                itemButton.previousButtons = nil
            end

            local frameX = (item:GetX()) * BTWQUESTS_CHAIN_ITEM_WIDTH
            local frameY = (item:GetY()) * BTWQUESTS_CHAIN_ITEM_HEIGHT

            local x = item:GetX()
            local y = item:GetY()
        
            itemButton:SetPoint("TOP", 0, -(y * 80) - (self:GetHeight()/2) + 24);
            itemButton:SetPoint("LEFT", 95 * x + 12, 0);
            -- itemButton:SetPoint("CENTER", self:GetScrollChild(), "TOPLEFT", frameX, -frameY)

            local connectionIndex = 1
            local connectionItem = item:GetConnection(connectionIndex)
            while connectionItem do
                if not connectionItem:GetSkip() then
                    local connectionButton = buttons[connectionItem:GetIndex()] or self.itemPool:Acquire();
                    buttons[connectionItem:GetIndex()] = connectionButton

                    -- Storing this so the connected items can check if they should be next
                    if connectionButton.previousButtons == nil then
                        connectionButton.previousButtons = {}
                    end
                    table.insert(connectionButton.previousButtons, itemButton)
                    
                    local lineContainer = itemButton.linePool:Acquire();
                    
                    lineContainer.Background:SetStartPoint("CENTER", itemButton);
                    lineContainer.Background:SetEndPoint("CENTER", connectionButton);
                    
                    lineContainer.Active:SetStartPoint("CENTER", itemButton);
                    lineContainer.Active:SetEndPoint("CENTER", connectionButton);
                    
                    lineContainer.Complete:SetStartPoint("CENTER", itemButton);
                    lineContainer.Complete:SetEndPoint("CENTER", connectionButton);
                    
                    -- lineContainer:SetAlpha(1)
                    
                    lineContainer.PulseAlpha:Stop()
                        
                    lineContainer.Active:Hide()
                    lineContainer.Complete:Hide()
                    
                    if itemButton.status == "complete" then
                        lineContainer.Complete:Show()
                    elseif itemButton.status == "active" then
                        -- lineContainer.Active:Show()
                    end
                    
                    lineContainer.connection = connectionButton
                    lineContainer:SetShown(itemButton:IsShown() and connectionItem:GetVisible());
                end
                
                connectionIndex = connectionIndex + 1
                connectionItem = item:GetConnection(connectionIndex)
            end

            if item:GetVisible() then
                if self.scrollTo == nil and itemButton.status ~= "complete" then
                    if item:IsAside() then
                        if self.scrollToButtonAside == nil then
                            self.scrollToButtonAside = itemButton
                        end
                    elseif self.scrollToButton == nil then
                        self.scrollToButton = itemButton
                    end
                elseif type(self.scrollTo) == "number" and index == self.scrollTo then
                    self.scrollToButton = itemButton
                elseif type(self.scrollTo) == "table" and item:Equals(self.scrollTo) then
                    self.scrollToButton = itemButton
                end
            end
        end

        index = index + 1
        item = chain:GetItem(index)
    end

    return chain:GetNumItems()
end
function BtWQuestsChainMixin:Update()
    self:UpdateChain(self.chainID)
end
function BtWQuestsChainMixin:UpdateChain(chainID)
    local character = self:GetCharacter()

    local buttons = self.ItemButtons[chainID]
    assert(buttons ~= nil, "Cannot update a chain that was never added")

    local chain = BtWQuestsDatabase:GetChain(chainID, character)

    local index = 1
    local item = chain:GetItem(index)
    while item do
        if not item:GetSkip() then
            local itemButton = buttons[index]
            assert(itemButton, "Chain item was never added in the first place")
            
            local status = itemButton.status

            itemButton:Update(item)

            -- Check if the item should be next
            itemButton.IsNextAnim:Stop()
            if itemButton.status == nil and itemButton.previousButtons ~= nil then
                local available = true
                for i,previous in ipairs(itemButton.previousButtons) do
                    if previous.status ~= "complete" then
                        available = false
                        break
                    end
                end

                if available then
                    itemButton.ActiveTexture:Show()
                    itemButton.IsNextAnim:Play()
                end

                itemButton.previousButtons = nil
            end

            local connectionIndex = 1
            local connectionItem = item:GetConnection(connectionIndex)
            while connectionItem do
                if not connectionItem:GetSkip() then
                    local connectionButton = buttons[connectionItem:GetIndex()]

                    -- Storing this so the connected items can check if they should be next
                    if connectionButton.previousButtons == nil then
                        connectionButton.previousButtons = {}
                    end
                    table.insert(connectionButton.previousButtons, itemButton)
                end
                
                connectionIndex = connectionIndex + 1
                connectionItem = item:GetConnection(connectionIndex)
            end
            
            for lineContainer in itemButton.linePool:EnumerateActive() do
                if itemButton.status == "complete" and status ~= "complete" then
                    lineContainer.DefaultToCompleteAnim:Play()
                end
                lineContainer:SetShown(itemButton:IsShown() and lineContainer.connection:IsShown())
            end
        end

        index = index + 1
        item = chain:GetItem(index)
    end
    
    return chain:GetNumItems()
end

-- [[ Navbar ]]
BtWQuestsNavBarMixin = {}
function BtWQuestsNavBarMixin:OnShow()
    if not self.Initialized then
        local homeData = {
            type = "expansion",
            name = HOME,
            OnClick = BtWQuestsNavBarButtonMixin.OnClick
        }
        self.dropDown = CreateFrame("Frame", dropDownName, self, "UIDropDownMenuTemplate")
        UIDropDownMenu_SetInitializeFunction(self.dropDown, NavBar_DropDown_Initialize)
        UIDropDownMenu_SetDisplayMode(self.dropDown, "MENU")
        NavBar_Initialize(self, "BtWQuestsNavBarButtonTemplate", homeData, self.home, self.overflow);
        self.Initialized = true
    end
end
function BtWQuestsNavBarMixin:SetExpansion(id)
    NavBar_Reset(self)
end
function BtWQuestsNavBarMixin:SetCategory(id)
    local character = BtWQuestsCharacters:GetCharacter(self:GetParent():GetCharacter())

    NavBar_Reset(self)
    self:AddCategoryButtonParents(id, character)
end
function BtWQuestsNavBarMixin:SetChain(id)
    local character = BtWQuestsCharacters:GetCharacter(self:GetParent():GetCharacter())

    NavBar_Reset(self)
    self:AddChainButtonParents(id, character)
end
function BtWQuestsNavBarMixin:AddCategoryButtonParents(id, character)
    local id = tonumber(id)
    local _, name, _, expansion, parent = BtWQuestsDatabase:GetCategoryByID(id, character)
    
    if parent then
        self:AddCategoryButtonParents(parent, character)
    end
    
    self:AddCategoryButton(id, name, parent == nil and {
        type = "expansion",
        id = expansion,
    } or {
        type = "category",
        id = parent,
    }, character)
end
function BtWQuestsNavBarMixin:AddChainButtonParents(id, character)
    local id = tonumber(id)
    local _, name, _, expansion, parent = BtWQuestsDatabase:GetChainByID(id, character)
    
    if parent then
        self:AddCategoryButtonParents(parent, character)
    end
    
    self:AddChainButton(id, name, parent == nil and {
        type = "expansion",
        id = expansion,
    } or {
        type = "category",
        id = parent,
    }, character)
end
function BtWQuestsNavBarMixin:AddCategoryButton(id, name, parent, character)
    local id = tonumber(id)
    local buttonData = {
        type = "category",
        id = id,
        parent = parent,
        name = name,
        OnClick = BtWQuestsNavBarButtonMixin.OnClick,
        listFunc = BtWQuestsNavBarButtonMixin.GetList,
    }
    NavBar_AddButton(self, buttonData);
end
function BtWQuestsNavBarMixin:AddChainButton(id, name, parent)
    local id = tonumber(id)
    local buttonData = {
        type = "chain",
        id = id,
        parent = parent,
        name = name,
        OnClick = BtWQuestsNavBarButtonMixin.OnClick,
        listFunc = BtWQuestsNavBarButtonMixin.GetList,
    }
    NavBar_AddButton(self, buttonData);
end
function BtWQuestsNavBarMixin:GoToItem(item)
    local BtWQuestsFrame = self:GetParent()

    if item.type == "expansion" then
        BtWQuestsFrame:SelectExpansion(BtWQuestsFrame:GetExpansion())
    elseif item.type == "category" then
        BtWQuestsFrame:SelectCategory(item.id)
    elseif item.type == "chain" then
        BtWQuestsFrame:SelectChain(item.id)
    end
end
function BtWQuestsNavBarMixin:GetCharacter()
    return BtWQuestsCharacters:GetCharacter(self:GetParent():GetCharacter())
end

BtWQuestsNavBarButtonMixin = {}
function BtWQuestsNavBarButtonMixin:GetList()
    local character = self:GetParent():GetCharacter()

    local item = self.data.parent
    if item == nil then
        return
    end

    local sisters = {}
    if item.type == "expansion" then
        local parent = item.id

        local index = 1
        local sisterType, sisterId, sisterName, hidden = BtWQuestsDatabase:GetExpansionItemByIndex(parent, index, character)
        while sisterId do
            if not hidden then
                table.insert(sisters, {
                    id = {
                        type = sisterType,
                        id = sisterId
                    },
                    text = sisterName,
                    func = function(button, item)
                        self:GetParent():GoToItem(item)
                    end,
                })
            end
            
            index = index + 1
            sisterType, sisterId, sisterName, hidden = BtWQuestsDatabase:GetExpansionItemByIndex(parent, index, character)
        end
    elseif item.type == "category" then
        local parent = item.id

        local index = 1
        local sisterType, sisterId, sisterName, hidden = BtWQuestsDatabase:GetCategoryItemByIndex(parent, index, character)
        while sisterId do
            if not hidden then
                table.insert(sisters, {
                    id = {
                        type = sisterType,
                        id = sisterId
                    },
                    text = sisterName,
                    func = function(button, item)
                        self:GetParent():GoToItem(item)
                    end,
                })
            end
            
            index = index + 1
            sisterType, sisterId, sisterName, hidden = BtWQuestsDatabase:GetCategoryItemByIndex(parent, index, character)
        end
    elseif item.type == "chain" then
    end
    return sisters
end
function BtWQuestsNavBarButtonMixin:OnClick()
    self:GetParent():GoToItem(self.data)
end

-- [[ Character Dropdown ]]
BtWQuestsCharacterDropDownMixin = {}
function BtWQuestsCharacterDropDownMixin:OnShow()
    if not self.Initialized then
        UIDropDownMenu_SetWidth(self, 156, 23 * 2);
		UIDropDownMenu_SetInitializeFunction(self, self.Initialize);
        -- UIDropDownMenu_Initialize(self, self.Initialize);
        self.xOffset = 8
        self.yOffset = 15

        self.Initialized = true
    end
end
function BtWQuestsCharacterDropDownMixin:Initialize()
    local function Select (button)
        self:GetParent():SelectCharacter(button.value)
    end

    local name = UnitName("player")
    local realm = GetRealmName()
    local player = name .. "-" .. realm

    local current = self:GetParent():GetCharacter()

    local info = UIDropDownMenu_CreateInfo();
    info.text = player
    info.value = player
    info.func = Select
    info.checked = player == current

    UIDropDownMenu_AddButton(info)  

    for character in BtWQuestsCharacters:pairs() do
        if character ~= player then
            info.text = character
            info.value = character
            info.func = Select
            info.checked = character == current
            UIDropDownMenu_AddButton(info)
        end
    end
end

-- [[ Expansion Dropdown ]]
BtWQuestsExpansionDropDownMixin = {}
function BtWQuestsExpansionDropDownMixin:OnShow()
    if not self.Initialized then
        UIDropDownMenu_SetWidth(self, 156);
        UIDropDownMenu_JustifyText(self, "LEFT");
		UIDropDownMenu_SetInitializeFunction(self, self.Initialize);
        -- UIDropDownMenu_Initialize(self, self.Initialize);

        self.Initialized = true
    end
end
function BtWQuestsExpansionDropDownMixin:Initialize()
    local function Select (button)
        self:GetParent():GetParent():SelectExpansion(button.value)
    end

    local info = UIDropDownMenu_CreateInfo();
    local current = self:GetParent():GetParent():GetExpansion();
    for i=0,LE_EXPANSION_LEVEL_CURRENT do
        if BtWQuestsDatabase:HasExpansion(i) then
            info.text = BtWQuestsDatabase:GetExpansionByID(i)
            info.value = i;
            info.func = Select
            info.checked = i == current;

            UIDropDownMenu_AddButton(info)
        elseif i == current then
            info.text = _G['BTWQUESTS_EXPANSION_NAME' .. i]
            info.value = i;
            info.func = Select
            info.checked = i == current;

            UIDropDownMenu_AddButton(info)
        end
    end
end

-- [[ Tooltip ]]
BtWQuestsTooltipMixin = {}
function BtWQuestsTooltipMixin:SetChain(chainID, character)
    local chainID = tonumber(chainID)
    
    local _, name, _, _, _, _, numPrerequisites = BtWQuestsDatabase:GetChainByID(chainID, character)
    
    self:ClearLines()
    self:AddDoubleLine(name)
    
    if BtWQuestsDatabase:IsChainActive(chainID, character) then
        self:AddLine(BTWQUESTS_QUEST_CHAIN_ACTIVE)
    end
    
    local addedPrerequisite
    local i = 1
	for index = 1, numPrerequisites do
        local name, visible, skip, completed = BtWQuestsDatabase:GetChainPrerequisiteByID(chainID, index, character);
        if visible and not skip then
            if not addedPrerequisite then
                self:AddLine(" ")
                self:AddLine(BTWQUESTS_TOOLTIP_PREREQUISITES)
                addedPrerequisite = true
            end

            if completed then
                self:AddLine(" - " .. name, 0.5, 0.5, 0.5)
            else
                self:AddLine(" - " .. name, 1, 1, 1)
            end
        end
	end
    
    self:Show();
end
-- Custom function for displaying an active quest showing completed requirements
function BtWQuestsTooltipMixin:SetActiveQuest(id, character)
    local id = tonumber(id)
    
    local _, name = BtWQuestsDatabase:GetQuestByID(id, character)
    local questLogIndex = GetQuestLogIndexByID(id)
    local _, objectiveText = GetQuestLogQuestText(questLogIndex);
    
    self:ClearLines()
    self:AddDoubleLine(name)
    
    self:AddLine(GREEN_FONT_COLOR_CODE..QUEST_TOOLTIP_ACTIVE..FONT_COLOR_CODE_CLOSE)

    if objectiveText then
        self:AddLine(" ")
        self:AddLine(objectiveText, 1, 1, 1)
    end
    
    local requiredMoney = GetQuestLogRequiredMoney(questLogIndex);
    local numRequirements = GetNumQuestLeaderBoards(questLogIndex);

    local addedTitle
	for index = 1,numRequirements do
        local name, _, completed = GetQuestLogLeaderBoard(index, questLogIndex);
        if name then
            if not addedTitle then
                self:AddLine(" ")
                self:AddLine(QUEST_TOOLTIP_REQUIREMENTS)
                addedTitle = true
            end

            if completed then
                self:AddLine(" - " .. name, 0.5, 0.5, 0.5)
            else
                self:AddLine(" - " .. name, 1, 1, 1)
            end
        end
    end
    
    self:Show();
end
-- Doing this because TSM and Auctioneer tainted GameTooltip.SetHyperlink without checking if their variables exist
local dummpGameTooltip = CreateFrame("GameTooltip", "BtWQuestsDummyTooltip", UIParent, "GameTooltipTemplate")
dummpGameTooltip:Hide()
function BtWQuestsTooltipMixin:SetHyperlink(link, character)
    local _, _, color, type, text, name = string.find(link, "|?c?f?f?(%x*)|?H?([^:]+):([^|]+)|?h?%[?([^%[%]]*)%]?|?h?|?r?")
    
    if type == "quest" and character:IsPlayer() then
        local _, _, id = string.find(text, "^(%d+)")
        if GetQuestLogIndexByID(id) > 0 then
            self:SetActiveQuest(id, character)
        else
            dummpGameTooltip.SetHyperlink(self, link)
        end
    elseif type == "btwquests" then
        local _, _, subtype, id = string.find(text, "^([^:]*):(%d+)")
        
        assert(subtype == "chain")
        
        self:SetChain(id, character)
    else
        dummpGameTooltip.SetHyperlink(self, link)
    end
end

-- [[ Compatibility functions ]]
-- These functions use the character from BtWQuestsFrame
function BtWQuests_GetQuestName(id)
    return BtWQuestsDatabase:GetQuestName(id)
end
function BtWQuests_IsQuestCompleted(id)
    return BtWQuestsCharacters:GetCharacter(BtWQuestsFrame:GetCharacter()):IsQuestCompleted(id)
end
function BtWQuests_IsQuestActive(id)
    return BtWQuestsCharacters:GetCharacter(BtWQuestsFrame:GetCharacter()):IsQuestCompleted(id)
end

function BtWQuests_GetChainName(id)
    return BtWQuestsDatabase:GetChainName(id, BtWQuestsCharacters:GetCharacter(BtWQuestsFrame:GetCharacter()))
end
function BtWQuests_IsChainCompleted(id)
    return BtWQuestsCharacters:GetCharacter(BtWQuestsFrame:GetCharacter()):IsChainCompleted(id)
end
function BtWQuests_IsChainActive(id)
    return BtWQuestsCharacters:GetCharacter(BtWQuestsFrame:GetCharacter()):IsChainCompleted(id)
end

function BtWQuests_GetCategoryName(id)
    return BtWQuestsDatabase:GetCategoryName(id, BtWQuestsCharacters:GetCharacter(BtWQuestsFrame:GetCharacter()))
end
function BtWQuests_IsCategoryCompleted(id)
    return BtWQuestsCharacters:GetCharacter(BtWQuestsFrame:GetCharacter()):IsCategoryCompleted(id)
end
function BtWQuests_IsCategoryActive(id)
    return BtWQuestsCharacters:GetCharacter(BtWQuestsFrame:GetCharacter()):IsCategoryCompleted(id)
end

-- [[ Waypoint ]]
function BtWQuests_AddWaypoint(mapId, x, y, name)
    if TomTom and TomTom.AddWaypoint then
        TomTom:AddWaypoint(mapId, x, y, {
            title = name,
        })
    end
end
function BtWQuests_ShowMapWithWaypoint(mapId, x, y, name)
    BtWQuests_AddWaypoint(mapId, x, y, name)

    if not IsModifiedClick("CHATLINK") then
        OpenQuestLog(mapId)
    end
end

-- [[ Quest To Chain ]]
function BtWQuests_AddOpenChainMenuItem(self, questID)
    local item = BtWQuestsDatabase:GetQuestItem(questID, BtWQuestsCharacters:GetPlayer())
    if item then
        local info = UIDropDownMenu_CreateInfo()
        info.text = BTWQUESTS_OPEN_QUEST_CHAIN
        info.func = function(self, questID, item)
            BtWQuestsFrame:SelectCharacter(UnitName("player"), GetRealmName())
            BtWQuestsFrame:SelectItem(item.item)
        end
        info.arg1 = questID
        info.arg2 = item
        info.notCheckable = 1;
        info.checked = false;
        info.noClickSound = 1;
        UIDropDownMenu_AddButton(info, UIDROPDOWN_MENU_LEVEL)
    end
end

-- [[ Slash Command ]]
SLASH_BTWQUESTS1 = "/btwquests"
SlashCmdList["BTWQUESTS"] = function(msg)
    if msg == "minimap" then
        BtWQuestsMinimapButton_Toggle()
    else
        ToggleFrame(BtWQuestsFrame)
    end
end 


-- [[ Hyperlink Handling ]]

local function ChatFrame_Filter(self, event, msg, ...)
	msg = msg:gsub("%[btwquests:([^:]+):(%d+):([^:]+):([^%]]+)%]","|c%3|Hbtwquests:%1:%2|h[%4]|h|r");
    
	return false, msg, ...;
end

local events = {
	"CHAT_MSG_SAY",
	"CHAT_MSG_YELL",
	"CHAT_MSG_EMOTE",
	"CHAT_MSG_GUILD",
	"CHAT_MSG_OFFICER",
	"CHAT_MSG_PARTY",
	"CHAT_MSG_PARTY_LEADER",
	"CHAT_MSG_RAID",
	"CHAT_MSG_RAID_LEADER",
	"CHAT_MSG_RAID_WARNING",
	"CHAT_MSG_BATTLEGROUND",
	"CHAT_MSG_BATTLEGROUND_LEADER",
	"CHAT_MSG_WHISPER",
	"CHAT_MSG_WHISPER_INFORM",
	"CHAT_MSG_BN_WHISPER",
	"CHAT_MSG_BN_WHISPER_INFORM",
	"CHAT_MSG_BN_CONVERSATION",
	"CHAT_MSG_BN_INLINE_TOAST_BROADCAST",
	"CHAT_MSG_BN_INLINE_TOAST_BROADCAST_INFORM",
    "CHAT_MSG_CHANNEL",
    "CHAT_MSG_COMMUNITIES_CHANNEL",
};

for i, event in ipairs(events) do
	ChatFrame_AddMessageEventFilter(event, ChatFrame_Filter);
end

local original_InsertLink = ChatEdit_InsertLink
function ChatEdit_InsertLink(link)
	if link and link:find("^|c[^|]+|Hbtwquests") then
        local color, text, name = link:match("^|c([^|]+)|H([^|]+)|h%[([^%]]+)%]|h|r");
        link = format("[%s:%s:%s]", text, color, name)
    end
    
    return original_InsertLink(link)
end

local original_SetHyperlink = ItemRefTooltip.SetHyperlink;
function ItemRefTooltip:SetHyperlink(link)
	if link:find("^btwquests") then
		if IsModifiedClick("CHATLINK") then
			ChatEdit_InsertLink("["..link.."]")
		else
            BtWQuestsFrame:SelectFromLink(link)
		end
	else
		original_SetHyperlink(self, link);
	end
end